# -*- coding: utf8 -*-

from __future__ import unicode_literals
from copy import deepcopy
from operator import and_

from django.core.exceptions import ValidationError
from django.db.models import Q
import six

from filternaut.tree import Tree, Leaf


class FilterTree(Tree):
    """
    FilterTrees instances are the result of ORing or ANDing Filter instances,
    or other FilterTree instances, together.

    FilterTree provides a simple API for dealing with a set of filters. They
    may also be iterated over to gain access to the Filter instances directly.
    """

    def __init__(self, negate=False, *args, **kwargs):
        self.negate = negate
        super(FilterTree, self).__init__(*args, **kwargs)

    def __invert__(self):
        inverted = deepcopy(self)
        inverted.negate = not self.negate
        return inverted

    @property
    def valid(self):
        """
        A boolean indicating whether all filters parsed successfully. Cannot be
        read before ``parse()`` has been called.
        """
        return not self.errors

    @property
    def errors(self):
        """
        A dictionary of errors met while parsing. The errors are keyed by their
        field's source value. Each key's value is a list of errors.

        ``errors`` cannot be read before ``parse()`` has been called.
        """
        errors = {}
        for filter in self:
            errors.update(filter.errors or {})
        return errors

    @property
    def Q(self):
        """
        A Django "Q" object, built by combining input data with the filter
        definitions. Cannot be read before ``parse()`` has been called.
        """
        qs = self.left.Q, self.right.Q

        # 'empty' Filters generate empty Q objects. These don't falsify other Q
        # objects to which they are ANDed, so we simulate that behaviour here.
        # In other words, Q(<real query>) & Q() == Q(<real query>), whereas we
        # want it to == Q().
        if self.operator == and_ and not all(qs):
            q = Q()
        else:
            q = self.operator(*qs)

        return ~q if self.negate else q

    def parse(self, data):
        for filter in self:
            filter.parse(data)


class Filter(Leaf):
    """
    A Filter instance builds a django.db.models.Q object by pulling a value
    from arbitrary native data, e.g. a set of query params.

    It can be ORed or ANDed together with other Filter instances in the same
    way Q objects can.
    """

    tree_class = FilterTree

    def __init__(self, dest, **kwargs):
        self.dest = dest
        self.source = kwargs.get('source', dest)
        self.lookups = kwargs.get('lookups', ['exact'])
        self.required = kwargs.get('required', False)
        self.negate = kwargs.get('negate', False)

        # None is a valid default -- consider exclude(groups=None) -- so use
        # the absence or presence of self.default to indicate whether a default
        # should be used.
        if 'default' in kwargs:
            self.default = kwargs['default']
            self.default_lookup = kwargs.get('default_lookup', 'exact')

        self._filters = {}
        self._errors = {}
        self.parsed = False

        # lazy folk can provide a single lookup as a string.
        if isinstance(self.lookups, six.string_types):
            self.lookups = [self.lookups, ]

    def __invert__(self):
        """
        Invert the sense of this filter.
        """
        inverted = deepcopy(self)
        inverted.negate = not self.negate
        return inverted

    def make_source_dest_pair(self, lookup):
        """
        Take a 'lookup' value, such as 'contains' or 'lte', and combine it with
        this field's source, returning e.g. username__contains.

        If lookup is None, the raw 'source' is returned.
        """
        if lookup in (None, 'exact'):
            source = self.source
            dest = self.dest
        else:
            source = '{}__{}'.format(self.source, lookup)
            dest = '{}__{}'.format(self.dest, lookup)
        return source, dest

    def parse(self, data):
        """
        Look through the provided dict-like data for keys which match this
        Filter's source.  This includes keys containg lookup affixes such as
        'contains' or 'lte'.

        Once this method has been called, the ``errors``, ``valid`` and ``Q``
        attributes become usable.
        """
        self._filters, self._errors = self.get_q_kwargs(data)
        self.parsed = True

    def clean(self, value):
        """
        Validate and normalise ``value`` for use in filtering. This
        implementation is a no-op; subclasses may do more work here.
        """
        return value

    def get_q_kwargs(self, data):
        """
        Look through ``data`` for keys which match any of the source /
        source__lookup values for this Filter. Any present are cleaned and kept
        for filtering. Errors are likewise kept.

        Returns a dictionary of filters keyed by dest (suitable for unpacking
        into a Q constructor) and a dictionary of errors, keyed by source.
        """
        filters, errors = {}, {}
        for lookup in self.lookups:
            source, dest = self.make_source_dest_pair(lookup)
            try:
                value = self.get_source_value(source, data)
                filters[dest] = self.clean(value)
            except KeyError:
                pass
            except ValidationError as ex:
                errors[source] = ex.messages

        # Use a default if there were no filters.
        if not filters and hasattr(self, 'default'):
            filters = self.make_default_filter()

        # If there are still no filters but this filter is required, that's an
        # error.
        if not filters and self.required:
            errors[self.source] = ['This field is required']

        return filters, errors

    def make_default_filter(self):
        """
        Construct a default filter dictionary to be used if no source data was
        found during parsing.
        """
        dest = '{}__{}'.format(self.dest, self.default_lookup)
        return {dest: self.default}

    def get_source_value(self, key, data):
        """
        Pull ``key`` from ``data``. This implementation is trivial, but
        subclasses may want to perform non-trivial inspections of ``data`` and
        can subclass here to do so.
        """
        return data[key]

    @property
    def valid(self):
        """
        A boolean indicating whether this Filter registered any errors during
        parsing. Raises a ValueError if ``parse()`` has not been called.
        """
        if not self.parsed:
            raise ValueError(
                "Must call parse() on filters before checking validity")
        return not self._errors

    @property
    def errors(self):
        """
        A dictionary of errors (keyed by source) listing any problems
        encountered during parsing. Typical entries include validation errors
        and failures to provide values where required.  Raises a ValueError if
        ``parse()`` has not been called.
        """
        if not self.parsed:
            raise ValueError(
                "Must call parse() on filters before reading errors")
        if not self._errors:
            return None
        else:
            return self._errors

    @property
    def Q(self):
        """
        A Django "Q" object, built by combining input data with this filter's
        definition. Cannot be read before ``parse()`` has been called.
        """
        if not self.parsed:
            raise ValueError(
                "Must call parse() on filters before using Q")
        q = Q(**self._filters)
        return ~q if self.negate else q



# -*- coding: utf8 -*-

from __future__ import unicode_literals, absolute_import

from django.forms import (
    CharField, IntegerField, DateField, TimeField, DateTimeField, RegexField,
    EmailField, ImageField, URLField, BooleanField, NullBooleanField,
    ChoiceField, MultipleChoiceField, ComboField, MultiValueField, FloatField,
    DecimalField, SplitDateTimeField, IPAddressField, FilePathField, SlugField,
    TypedChoiceField)

# note GenericIPAddressField and TypedMultipleChoiceField are conditionally
# imported later in this file; they are not available in all versions of
# Django.

from filternaut import Filter

__all__ = [
    'BooleanFilter', 'CharFilter', 'ChoiceFilter', 'ComboFilter', 'DateFilter',
    'DateTimeFilter', 'DecimalFilter', 'EmailFilter', 'FilePathFilter',
    'FloatFilter', 'IPAddressFilter', 'ImageFilter', 'FieldFilter',
    'IntegerFilter', 'MultiValueFilter', 'MultipleChoiceFilter',
    'NullBooleanFilter', 'RegexFilter', 'SlugFilter', 'SplitDateTimeFilter',
    'TimeFilter', 'TypedChoiceFilter', 'URLFilter']


class FieldFilter(Filter):
    """
    FieldFilters use a django.forms.Field to clean their input value when
    generating a Q object.

    This class is designed to be extended by subclasses which provide their own
    form-field instances. However, you could use it in combination with a
    custom field like so:

        filter = FieldFilter(SpecialField(), dest='...')
    """

    def __init__(self, dest, field, **kwargs):
        self.field = field
        super(FieldFilter, self).__init__(dest, **kwargs)

    def clean(self, value):
        return self.field.clean(value)


# -- mixtures of fieldfilter and django fields requiring additional arguments


class ChoiceFilter(FieldFilter):
    def __init__(self, dest, choices, *args, **kwargs):
        field = ChoiceField(choices=choices)
        super(ChoiceFilter, self).__init__(dest, field=field, *args, **kwargs)


class RegexFilter(FieldFilter):
    def __init__(self, dest, regex, *args, **kwargs):
        field = RegexField(regex=regex)
        super(RegexFilter, self).__init__(dest, field=field, *args, **kwargs)


class FilePathFilter(FieldFilter):
    def __init__(self, dest, path, *args, **kwargs):
        field = FilePathField(path=path)
        super(FilePathFilter, self).__init__(dest, field=field,
                                             *args, **kwargs)


# -- simple mixtures of fieldfilter and django fields.

class FieldMixin(object):
    def __init__(self, dest, **kwargs):
        field = self.field_class()
        super(FieldMixin, self).__init__(dest, field=field, **kwargs)


class CharFilter(FieldMixin, FieldFilter):
    field_class = CharField


class BooleanFilter(FieldMixin, FieldFilter):
    field_class = BooleanField


class ComboFilter(FieldMixin, FieldFilter):
    field_class = ComboField


class DateFilter(FieldMixin, FieldFilter):
    field_class = DateField


class DateTimeFilter(FieldMixin, FieldFilter):
    field_class = DateTimeField


class DecimalFilter(FieldMixin, FieldFilter):
    field_class = DecimalField


class EmailFilter(FieldMixin, FieldFilter):
    field_class = EmailField


class FloatFilter(FieldMixin, FieldFilter):
    field_class = FloatField


class IPAddressFilter(FieldMixin, FieldFilter):
    field_class = IPAddressField


class ImageFilter(FieldMixin, FieldFilter):
    field_class = ImageField


class IntegerFilter(FieldMixin, FieldFilter):
    field_class = IntegerField


class MultiValueFilter(FieldMixin, FieldFilter):
    field_class = MultiValueField


class MultipleChoiceFilter(FieldMixin, FieldFilter):
    field_class = MultipleChoiceField


class NullBooleanFilter(FieldMixin, FieldFilter):
    field_class = NullBooleanField


class SlugFilter(FieldMixin, FieldFilter):
    field_class = SlugField


class SplitDateTimeFilter(FieldMixin, FieldFilter):
    field_class = SplitDateTimeField


class TimeFilter(FieldMixin, FieldFilter):
    field_class = TimeField


class TypedChoiceFilter(FieldMixin, FieldFilter):
    field_class = TypedChoiceField


class URLFilter(FieldMixin, FieldFilter):
    field_class = URLField


# -- mixtures of fieldfilter and django fields which were not present in older
# -- django versions

try:
    # django 1.3 and older did not have GenericIPAddressField or
    # TypedMultipleChoiceField, so we must conditionally define our matching
    # Filters.
    from django.forms import GenericIPAddressField, TypedMultipleChoiceField
except ImportError:
    pass
else:
    class GenericIPAddressFilter(FieldMixin, FieldFilter):
        field_class = GenericIPAddressField

    class TypedMultipleChoiceFilter(FieldMixin, FieldFilter):
        field_class = TypedMultipleChoiceField

    __all__.extend((
        'GenericIPAddressFilter',
        'TypedMultipleChoiceFilter'))



from setuptools import setup

setup(
    name='django-filternaut',
    version='0.0.2',
    author='Ben Graham',
    author_email='bgraham@bgraham.com.au',
    description=('Construct arbitrarily complex Django "Q" filters from flat '
                 'data, such as query parameters.'),
    long_description=open('README.rst').read(),
    url='http://github.com/bennullgraham/django-filternaut/',
    license='BSD',
    packages=['filternaut'],
    install_requires=['six==1.9.0'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Utilities',
        'Framework :: Django',
    ],
)



# -*- coding: utf8 -*-

from __future__ import unicode_literals, print_function, absolute_import
from operator import or_, and_

"""
This module ends up looking similar to django.utils.tree. It has a simpler
design allowed by the simpler use case. In particular there are
only ever two children, `left' and `right', which in turn allows the 'operator'
(what django terms a 'connector') to be literally a function from stdlib's
operator.

Finally, there are separate leaf and tree classes. This increases sanity when
these classes are extended in filternaut.filters.
"""


class Tree(object):

    def __init__(self, left, right, operator):
        assert operator in (or_, and_)
        self.left = left
        self.right = right
        self.operator = operator

    def __iter__(self):
        for l in self.left:
            yield l
        for r in self.right:
            yield r

    def __and__(self, other):
        return self.__class__(left=self, right=other, operator=and_)

    def __or__(self, other):
        return self.__class__(left=self, right=other, operator=or_)


class Leaf(object):

    # subclasses could specify a tree-subclass here. e.g. MyLeaf.tree_class =
    # MyTree.
    tree_class = Tree

    def __iter__(self):
        yield self

    def __and__(self, other):
        return self.tree_class(left=self, right=other, operator=and_)

    def __or__(self, other):
        return self.tree_class(left=self, right=other, operator=or_)
